/*
 * @(#)%Name% %Version% %Time%
 * 
 * Copyright (c) 1997 Rutgers University
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that (i) the above copyright notice and this permission notice
 * appear in all copies of the software and related documentation. (ii) 
 * the name of Rutgers University may not be used in any advertising
 * or publicity relating to the software without the specific, prior
 * written permission of Rutgers University.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL RUTGERS UNIVERSITY BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY
 * THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOFTWARE.
 * 
 * CopyrightVersion %Version%
 * @author		%Author%
 * @version		%Version%, %Time%
 * @since		ACT %Version%
 *
 */

package EDU.rutgers.CORBA.GIOP.messages;

import java.io.ByteArrayInputStream;
import java.io.InputStream;

import ORG.omg.CORBA.OctetSequence;
import ORG.omg.CORBA.EMPTYByteArray;

import EDU.rutgers.act.presentation.TypeReader;
import EDU.rutgers.act.presentation.TypeWriter;

import EDU.rutgers.CORBA.presentation.*;

import EDU.rutgers.CORBA.IOP.IOR;

/**
 * [CORBA 2.0 Specification 12-22]
 *
 * @see GIOPMessage
 *
 * @see GIOPRequestMessage
 * @see GIOPReplyMessage
 * @see GIOPCancelRequestMessage
 * @see GIOPLocateRequestMessage
 * @see GIOPCloseConnectionMessage
 * @see GIOPMessageErrorMessage
 *
 * @since ACT 0.1
 */
public final class GIOPLocateReplyMessage extends GIOPMessage
{
    /**
     * Used to determine whether a LocateReply body exist.
     * Only OBJECT_FORWARD type of message contains the body
     *
     * @see LocateStatusType
     *
     * @since ACT 0.1
     */
	public LocateStatusType locate_status_;

    private byte[] locate_reply_body_ = EMPTYByteArray.EMPTY; 

    private int locateReplyHeaderSize_; // network size

    /**
     * Empty constructor, it could only be used by the GIOPMessageFactory
     *
     * @since ACT 0.1
     */
	protected GIOPLocateReplyMessage() {
        super(MessageType.LocateReply);
    }

    /**
     * @param int request_id
     * @param LocateStatusType locate status
     * 
     * @since ACT 0.1
     */
    public GIOPLocateReplyMessage(
        int request_id,
        LocateStatusType locate_status) {
        super(MessageType.LocateReply);

        request_id_ = request_id;
        locate_status_ = locate_status;

        calLocateReplyHeaderSize();
    }

    /**
     * Calculate the header network size. In case the user just resuse this object
     * by replacing some member fields, this method forces to reset the GIOP header
     * message_size field.
     *
     * @since ACT 0.1
     */
    public void calLocateReplyHeaderSize() {
        int twoHeadersSize = super.giopHeaderSize_;

        twoHeadersSize += super.ts_.calAlignGap(twoHeadersSize, 4) + 4; // request_id_
        twoHeadersSize += super.ts_.calAlignGap(twoHeadersSize, 4) + 4; // locate_status_

        locateReplyHeaderSize_ = twoHeadersSize - super.giopHeaderSize_;
        
        super.setMessageSize(locateReplyHeaderSize_ + locate_reply_body_.length);
    }

    /**
     * Must be called before setting any LocateReply body.
     * The user should use this value to determine the LocateReply body alignment.
     * This alignment is based on the GIOPMessage.ts_ transfer syntax.
     *
     * @return int size of the GIOP MessageHeader and LocateReplyHeader 
     * after the corresponding transfer syntax
     *
     * @see GIOPMessage#ts_
     * @see #setLocateReplyBody
     * 
     * @since ACT 0.1
     */
    public int getBeforeBodySize() {
        return locateReplyHeaderSize_ + super.giopHeaderSize_;
    }

    /**
     * Set the LocateReply body. If the locate_status is not OBJECT_FROWARD,
     * it just return. If true, this method would set the message_size also.
     * Normally, the user should first call getBeforeBodySize to get all headers size,
     * then format the body with that size as the index.
     *
     * <p>Here is the sample code to compose a LocateReplyMessage with locate_status is OBJECT_FORWARD
     *
     * <pre>
     * ...
     * // create the locate reply message
     * GIOPLocateReplyMessage locateReplyMsg = new GIOPLocateReplyMessage(
     *      locateRequestMsg.request_id_, LocateStatusType.OBJECT_FORWARD);
     *
     * // create the steam and writer
     * ByteArrayOutputStream bodyStream = new ByteArrayOutputStream();
     * TypeWriter bodyWriter = new CDRTypeWriter(
     *                               new OutputStreamWriter(bodyStream));
     *               
     * // prepare the ProfileBody
     * ProfileBody profileBody = new ProfileBody("biko", (short)9000, requestMsg.object_key_);
     * profileBody.write(bodyWriter);
     * bodyWriter.flush(); // finished
     * byte[] profileEncapsulation = bodyStream.toByteArray();
     *
     * // discard the ProfileBody encapsulation
     * // so the stream and writer could be resued again
     * bodyStream.reset(); 
     *
     * // put the profile body into the TaggedProfile
     * TaggedProfile taggedProfile = new TaggedProfile(TAG_INTERNET_IOP.value, profileEncapsulation);
     *
     * // Prepare the IOR
     * TaggedProfile[] profiles = new TaggedProfile[1];
     * profiles[0] = taggedProfile;
     * IOR ior = new IOR("Chinese Weicong Wang", profiles);
     *
     * // reset the writer
     * bodyWriter.setByteOrder(locateReplyMsg.byte_order_);
     * bodyWriter.setIndex(locateReplyMsg.getBeforeBodySize());
     *
     * ior.write(bodyWriter);
     * bodyWriter.flush(); // finished
     * byte[] locateReplyBody = bodyStream.toByteArray();
     *                
     * // set the reply body
     * locateReplyMsg.setLocateReplyBody(locateReplyBody);
     * ...
     * </pre>
     *
     * @param byte[] locate reply body
     *
     * @see #getBeforeBodySize
     * 
     * @since ACT 0.1
     */
    public void setLocateReplyBody(byte[] locate_reply_body) {
        if (locate_status_ == LocateStatusType.OBJECT_FORWARD && 
            locate_reply_body_ != null) {
            locate_reply_body_ = locate_reply_body;
            super.setMessageSize(locateReplyHeaderSize_ + locate_reply_body_.length);
        }
    }

    /**
     * Retrieve the locate reply body. This body is only byte array and it is 
     * aligned at the beginning of the GIOP MessageHeader.
     *
     * @return byte[] body
     *
     * @see #setLocateReplyBody
     *
     * @since ACT 0.1
     */
    public byte[] getLocateReplyBody() {
        return locate_reply_body_;
    }

    /**
     * Retrieve the IOR from the locate reply message if the locate status is OBJECT_FORWARD.
     * From the IOR, the user could call its getIIOPProfile to get the IIOP ProfileBody.
     * From the IIOP ProfileBody, the user could get the host name, port, and object_key
     *
     * @return IOR If wrong reply status and any error happen, return null.
     *
     * @see EDU.rutgers.act.IOP.IOR;
     * @see EDU.rutgers.act.IOP.IOR#getIIOPProfile
     * @see EDU.rutgers.act.IIOP.ProfileBody
     *
     * @since ACT 0.1
     */
    public IOR getIOR() {
        if (locate_status_ != LocateStatusType.OBJECT_FORWARD) {
            return null;
        }
        ByteArrayInputStream stream = new ByteArrayInputStream(locate_reply_body_);
        TypeReader in = new CDRTypeReader(stream);

        in.setIndex(super.giopHeaderSize_ + locateReplyHeaderSize_);
        in.setByteOrder(super.byte_order_);

        IOR ior = new IOR();
        try {
            ior.read(in);
        } catch (java.io.IOException e) {
            // what should we do here ?
        }
        return ior;
    }

    /**
     * Read from the TypeReader stream
     *
     * @exception java.io.IOException If any IO error occurs
     *
     * @see ORG.omg.CORBA.Streamable#read
     * @see EDU.rutgers.act.presentation.TypeReader
     *
     * @since ACT 0.1
     */
    public void read(TypeReader in) 
		throws java.io.IOException {
        super.read(in);

		request_id_ = in.readULong();
		locate_status_ = LocateStatusType.type(in.readULong());

        if (locate_status_ == LocateStatusType.OBJECT_FORWARD) {
            int twoHeadersSize = in.getIndex();
            locateReplyHeaderSize_ = twoHeadersSize - super.giopHeaderSize_;
            int bodySize = super.message_size_ - locateReplyHeaderSize_;

            locate_reply_body_ = new byte[bodySize];
            in.readOctetArray(locate_reply_body_, 0, bodySize);
        }
	}

    /**
     * Write to the TypeWriter stream.
     *
     * @exception java.io.IOException If any IO error occurs
     *
     * @see ORG.omg.CORBA.Streamable#write
     * @see EDU.rutgers.act.presentation.TypeWriter
     *
     * @since ACT 0.1
     */
    public void write(TypeWriter out)
		throws java.io.IOException {
        super.write(out);

		out.writeLong(request_id_);
		out.writeLong(locate_status_.value());

        if (locate_status_ == LocateStatusType.OBJECT_FORWARD) {
            out.writeOctetArray(locate_reply_body_, 0, locate_reply_body_.length);
        }
	}

// Object
/*
    public boolean equals(Object o) {
		if(super.equals(o) && (o != null) && (o instanceof GIOPLocateReplyMessage)) {
            if (o == this) {
                return true;
            }
            GIOPLocateReplyMessage msg = (GIOPLocateReplyMessage)o;
            if (request_id_ == msg.request_id_ && 
                locate_status_ == msg.locate_status_) {
                return true;
            }
        }
        return false;
    }
*/

    /**
     * Returns a string representation. 
     *
     * @return String description
     *
     * @since ACT 0.1
     */
    public String toString() {
        StringBuffer buf = new StringBuffer(super.toString()).
            append("\nGIOP:LocateReplyMessage").
            append("\n\t[Header] ").
            append("\n\trequest_id: ").append(request_id_).
            append("\n\tlocate_status: ").append(locate_status_.toString()).
            append("\n\t[Body] ").append(OctetSequence.byteArray2String(locate_reply_body_));
        return String.valueOf(buf);
	}
}
