/*
 * @(#)%Name% %Version% %Time%
 * 
 * Copyright (c) 1997 Rutgers University
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that (i) the above copyright notice and this permission notice
 * appear in all copies of the software and related documentation. (ii) 
 * the name of Rutgers University may not be used in any advertising
 * or publicity relating to the software without the specific, prior
 * written permission of Rutgers University.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL RUTGERS UNIVERSITY BE LIABLE FOR ANY SPECIAL,
 * INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY
 * THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOFTWARE.
 * 
 * CopyrightVersion %Version%
 * @author		%Author%
 * @version		%Version%, %Time%
 * @since		ACT %Version%
 *
 */

// last modified 01/20/98

package EDU.rutgers.CORBA.GIOP.messages;

import java.io.ByteArrayInputStream;
import java.io.InputStream;

import ORG.omg.CORBA.OctetSequence;
import ORG.omg.CORBA.EMPTYByteArray;

import EDU.rutgers.act.presentation.TypeReader;
import EDU.rutgers.act.presentation.TypeWriter;

import EDU.rutgers.CORBA.presentation.*;

import EDU.rutgers.CORBA.IOP.ServiceContext;
import EDU.rutgers.CORBA.IOP.IOR;

/**
 * [CORBA 2.0 Specification 12-18]
 *
 * @see GIOPReplyMessage
 *
 * @see GIOPRequestMessage
 * @see GIOPCancelRequestMessage
 * @see GIOPLocateRequestMessage
 * @see GIOPLocateReplyMessage
 * @see GIOPCloseConnectionMessage
 * @see GIOPMessageErrorMessage
 *
 * @since ACT 0.1
 */
public final class GIOPReplyMessage extends GIOPMessage
{
    /**
     * Contains ORB service data being passed from the server to the client
     *
     * @see EDU.rutgers.act.IOP.ServiceContext
     *
     * @since ACT 0.1
     */
    public ServiceContext[] service_context_;

    /**
     * Indicates the completion status of the associated request and also 
     * determines part of the reply body content.
     *
     * @see ReplyStatusType 
     *
     * @since ACT 0.1
     */
    public ReplyStatusType reply_status_;

    private byte[] reply_body_ = EMPTYByteArray.EMPTY;

    private int replyHeaderSize_;

    /**
     * Empty constructor, it could only be used by the GIOPMessageFactory
     *
     * @since ACT 0.1
     */
	protected GIOPReplyMessage() {
        super(MessageType.Reply);
    }

    /**
     * @param ServiceContext[] service context
     * @param request_id request id
     * @param ReplyStatusType reply status
     *
     * @see EDU.rutgers.act.IOP.ServiceContext
     * @see #request_id_
     * @see ReplyStatusType 
     *
     * @since ACT 0.1
     */
    public GIOPReplyMessage(
		ServiceContext[] service_context,
        int request_id,
        ReplyStatusType reply_status) {
        super(MessageType.Reply);

        service_context_ = service_context;
        request_id_ = request_id;
        reply_status_ = reply_status;

        calReplyHeaderSize();
    }

    /**
     * Calculate the header network size. In case the user just resuse this object
     * by replacing some member fields, this method forces to reset the GIOP header
     * message_size field.
     *
     * @since ACT 0.1
     */
    public void calReplyHeaderSize() {
        int twoHeadersSize = super.giopHeaderSize_;

        // service_context_[].length as unsigned long
        twoHeadersSize += super.ts_.calAlignGap(twoHeadersSize, 4) + 4; 
		int len = service_context_.length;
		for(int i = len; i-- != 0;) {
			// service_context_.context_id and service_context_.length as unsigned long
            twoHeadersSize += (super.ts_.calAlignGap(twoHeadersSize, 4) + 4) << 1;
            // service_context_.context_data_ as byte array
            len = service_context_[len - i - 1].context_data_.length;
            twoHeadersSize += super.ts_.calAlignGap(twoHeadersSize, 1) + len; 
		}
        twoHeadersSize += super.ts_.calAlignGap(twoHeadersSize, 4) + 4; // request_id
        twoHeadersSize += super.ts_.calAlignGap(twoHeadersSize, 4) + 4; // reply_status
        replyHeaderSize_ = twoHeadersSize - super.giopHeaderSize_;

        super.setMessageSize(replyHeaderSize_ + reply_body_.length);
    }

    /**
     * Must be called before setting any Reply body.
     * The user should use this value to determine the Reply body alignment.
     * This alignment is based on the GIOPMessage.ts_ transfer syntax.
     *
     * @return int size of the GIOP MessageHeader and LocateReplyHeader 
     * after the corresponding transfer syntax
     *
     * @see GIOPMessage#ts_
     * @see #setReplyBody
     * 
     * @since ACT 0.1
     */
    public int getBeforeBodySize() {
        return replyHeaderSize_ + super.giopHeaderSize_;
    }


    /**
     * Set the Reply body.
     * Normally, the user should first call getBeforeBodySize to get all headers size,
     * then format the body with that size as the index.
     *
     * In the GIOPLocateReplyMessage, there is somce sample code about how to compose
     * the IOR.
     *
     * <p>Here is the sample code to compose a ReplyMessage.
     *
     * <pre>
     * ...
     * // create the locate reply message
     * GIOPReplyMessage replyMsg = new GIOPReplyMessage(...);
     *
     * // create the steam and writer
     * ByteArrayOutputStream bodyStream = new ByteArrayOutputStream();
     * TypeWriter bodyWriter = new CDRTypeWriter(
     *                               new OutputStreamWriter(bodyStream));
     *               
     * // reset the writer
     * bodyWriter.setByteOrder(replyMsg.byte_order_);
     * bodyWriter.setIndex(replyMsg.getBeforeBodySize());
     *
     * // write the necessary fields into the TypeWriter
     * ...
     * bodyWriter.flush(); // finished
     * byte[] replyBody = bodyStream.toByteArray();
     *                
     * // set the reply body
     * replyMsg.setReplyBody(replyBody);
     * ...
     * </pre>
     *
     * @param byte[] reply body
     *
     * @see GIOPLocateReplyMessage#setLocateReplyBody
     * @see #getBeforeBodySize
     * @see #getReplyBody
     * 
     * @since ACT 0.1
     */
    public void setReplyBody(byte[] reply_body) {
        if (reply_body == null) {
            return;
        }
        reply_body_ = reply_body;
        super.setMessageSize(replyHeaderSize_ + reply_body_.length);
    }

    /**
     * Retrieve the reply body. This body is only byte array and it is 
     * aligned at the beginning of the GIOP MessageHeader.
     *
     * @return byte[] body
     *
     * @see #setReplyBody
     *
     * @since ACT 0.1
     */
    public byte[] getReplyBody() {
        return reply_body_;
    }

    /**
     * Retrieve the IOR from the reply message if the locate status is LOCATION_FORWARD.
     * From the IOR, the user could call its getIIOPProfile to get the IIOP ProfileBody.
     * From the IIOP ProfileBody, the user could get the host name, port, and object_key.
     *
     * @return IOR If wrong reply status and any error happen, return null.
     *
     * @see EDU.rutgers.act.IOP.IOR;
     * @see EDU.rutgers.act.IOP.IOR#getIIOPProfile
     * @see EDU.rutgers.act.IIOP.ProfileBody
     *
     * @since ACT 0.1
     */
    public IOR getIOR() {
        if (reply_status_ != ReplyStatusType.LOCATION_FORWARD) {
            return null;
        }
        ByteArrayInputStream stream = new ByteArrayInputStream(reply_body_);
        TypeReader in = new CDRTypeReader(stream);

        in.setIndex(super.giopHeaderSize_ + replyHeaderSize_);
        in.setByteOrder(super.byte_order_);

        IOR ior = new IOR();
        try {
            ior.read(in);
        } catch (java.io.IOException e) {
            ior = null;
            // what should we do here ?
        }
        return ior;
    }

    /**
     * Read from the TypeReader stream
     *
     * @exception java.io.IOException If any IO error occurs
     *
     * @see ORG.omg.CORBA.Streamable#read
     * @see EDU.rutgers.act.presentation.TypeReader
     *
     * @since ACT 0.1
     */
    public void read(TypeReader in) 
		throws java.io.IOException {
        // GIOP header
        super.read(in);

        // reply header
		int len = in.readULong();
		service_context_ = new ServiceContext[len];
		for(int i = len; i-- != 0;) {
            service_context_[len - i - 1] = new ServiceContext();
			service_context_[len - i - 1].read(in);
		}	
		request_id_ = in.readULong();
		reply_status_ = ReplyStatusType.type(in.readULong());

        // reply body
        int twoHeadersSize = in.getIndex();
        replyHeaderSize_ = twoHeadersSize - super.giopHeaderSize_;
        int bodySize = super.message_size_ - replyHeaderSize_;

        reply_body_ = new byte[bodySize];
        in.readOctetArray(reply_body_, 0, bodySize);
    }

    /**
     * Write to the TypeWriter stream.
     *
     * @exception java.io.IOException If any IO error occurs
     *
     * @see ORG.omg.CORBA.Streamable#read
     * @see EDU.rutgers.act.presentation.TypeWriter
     *
     * @since ACT 0.1
     */
    public void write(TypeWriter out)
		throws java.io.IOException {
        // GIOP header
        super.write(out);

        // reply header
		int len = service_context_.length;
		out.writeULong(len);
		for(int i = len; i-- != 0;) {
			service_context_[len - i - 1].write(out);
		}	
		out.writeULong(request_id_);
		out.writeULong(reply_status_.value());

        // reply body
        out.writeOctetArray(reply_body_, 0, reply_body_.length);
	}

// Object
/*
    public boolean equals(Object o) {
		if(super.equals(o) && (o != null) && (o instanceof GIOPReplyMessage)) {
            if (o == this) {
                return true;
            }
            GIOPReplyMessage msg = (GIOPReplyMessage)o;
            int len = service_context_.length;
            if (len != msg.service_context_.length) {
                return false;
            }
		    for (int i = len; i-- != 0; ) {
                if (service_context_[len - i].equals(msg.service_context_[len - i]) == false) {
                    return false;
                }
		    }
            if (request_id_ == msg.request_id_ && reply_status_ == msg.reply_status_) {
                return true;
            }
        }
        return false;
    }
*/

    /**
     * Returns a string representation of the GIOPReplyMessage. 
     *
     * @return String GIOPReplyMessage description
     *
     * @since ACT 0.1
     */
    public String toString() {
        StringBuffer buf = new StringBuffer(super.toString()).
            append("\nGIOP:GIOPReplyMessage").
            append("\n\t[Header]");
        int len = service_context_.length;
		for(int i = len; i-- != 0;) {
			buf.append("\n\t").append(service_context_[len - i - 1].toString());
		}
		buf.append("\n\trequest_id: ").append(request_id_).
            append("\n\treply_status: ").append(reply_status_.toString()).
            append("\n\t[Body] ").append(OctetSequence.byteArray2String(reply_body_));
		return String.valueOf(buf);
    }
}
